import * as TE from "fp-ts/lib/TaskEither";
import * as IO from "fp-ts/lib/IO";
import * as RA from "fp-ts/lib/ReadonlyArray";
import * as RM from "fp-ts/lib/ReadonlyMap";
import * as T from "fp-ts/Task";
import * as S from "fp-ts/string";
import { pipe, identity } from "fp-ts/lib/function";
import type { Separated } from "fp-ts/lib/Separated";
import type { Magma } from "fp-ts/lib/Magma";
import { intercalate } from "fp-ts/lib/Foldable";
import { nextSchedule, type Schedule } from "../canary";
import { ConsoleLogger, type Logger } from "../util";

export interface Job {
  id: string;
  toString: () => string;
  execute: <TResult>() => TE.TaskEither<Error, TResult>;
  schedule: Schedule;
  maxRetries: number;
}

export const logExecutingJobs = (jobs: ReadonlyArray<Job>, now: Date, logger: Logger = ConsoleLogger) => {
  const stringifieds = pipe(
    jobs,
    RA.map(({ toString }) => toString()),
    (stringified) => intercalate(S.Monoid, RA.Foldable)("|", stringified)
  );
  return logger.log(`Executing ${stringifieds} at ${now.toUTCString()}`);
};

export const execute = <TResult>(
  jobs: ReadonlyArray<Job>
): T.Task<Separated<ReadonlyArray<[Job, Error]>, ReadonlyArray<[Job, TResult]>>> =>
  pipe(
    T.of(jobs),
    T.map(
      RA.map((job) =>
        pipe(
          job.execute(),
          TE.bimap(
            (error) => [job, error] as [Job, Error],
            (result) => [job, result] as [Job, TResult]
          )
        )
      )
    ),
    T.flatMap(RA.wilt(T.ApplicativePar)(identity))
  );

export type JobsTableRow = {
  scheduled: Date;
  retries: number;
};
export type JobsTable = ReadonlyMap<string, JobsTableRow>;

export const jobsTableRowMagma: Magma<JobsTableRow> = {
  concat: (a: JobsTableRow, b: JobsTableRow): JobsTableRow => {
    if (a.scheduled <= b.scheduled) {
      // always prefer the later schedule.
      return b;
    }
    //    if (a.retries <= b.retries) {
    //      return a;
    //    }
    return b;
  }
};

export const formatJobResults = <TResult>(
  jobsTable: JobsTable,
  { left, right }: Separated<ReadonlyArray<[Job, Error]>, ReadonlyArray<[Job, TResult]>>
) => {
  const failures = left
    .map(([job, err]) => ({
      job,
      retries: jobsTable.get(job.id)?.retries ?? job.maxRetries,
      err
    }))
    .filter(({ job, retries }) => retries >= job.maxRetries);
  const retries = left
    .map(([job, err]) => ({
      job,
      retries: jobsTable.get(job.id)?.retries ?? 0,
      err
    }))
    .filter(({ job, retries }) => retries < job.maxRetries);

  const failureMessages = failures
    .map(({ job, err, retries }) => job.toString() + ` | ${retries} / ${job.maxRetries} |  (err) :(  | ${err.message}`)
    .join("\n");
  const retryMessages = retries
    .map(
      ({ job, err, retries }) => job.toString() + ` | ${retries} / ${job.maxRetries} |  (retry) :/  | ${err.message}`
    )
    .join("\n");
  const successMessages = right.map(([job, val]) => job.toString() + " | (success) :) | " + val).join("\n");
  return `FAILURES:\n${failureMessages}\nRETRIES:\n${retryMessages}\nSUCCESSES:\n${successMessages}`;
};

export const perform = (
  jobsTable: JobsTable,
  jobs: ReadonlyArray<Job>,
  publishers: ReadonlyArray<(message: string) => TE.TaskEither<Error, number>>,
  lastPingAck: Date,
  publishAckEvery = 24 * (60 * 60 * 1_000),
  logger = ConsoleLogger
) =>
  pipe(
    T.of(jobs),
    T.bindTo("jobs"),
    T.bind(
      "now",
      T.fromIOK(() => IO.of(new Date()))
    ),
    T.bind("toExecute", ({ jobs, now }) =>
      pipe(
        jobs,
        RA.filter((job) => (jobsTable.get(job.id)?.scheduled ?? now) <= now),
        T.of
      )
    ),
    T.tap(({ toExecute, now }) => T.fromIO(logExecutingJobs(toExecute, now))),
    T.bind("results", ({ toExecute }) => execute(toExecute)),
    T.bind("shouldAck", ({ now }) => T.of(now.getTime() - lastPingAck.getTime() >= publishAckEvery)),
    T.tap(({ results, shouldAck }) => {
      if (results.left.length === 0 && !shouldAck) {
        return pipe(
          T.of({ right: [] as ReadonlyArray<number> }),
          T.tap(() => T.fromIO(logger.log("not publishing")))
        );
      }

      return pipe(
        publishers,
        RA.map((publish) => publish(formatJobResults(jobsTable, results))),
        RA.wilt(T.ApplicativePar)(identity),
        T.tap(({ left }) =>
          pipe(
            left
              ? logger.error("Encountered publishing errors: " + left.toString())
              : logger.log("Published successfully"),
            T.fromIO
          )
        )
      );
    }),
    T.map(({ now, results: { left, right }, shouldAck }) => {
      const ack = shouldAck ? now : lastPingAck;

      const leftResults = pipe(
        left,
        RA.map(([job]): readonly [string, JobsTableRow] => {
          const row = jobsTable.get(job.id);
          if (!row) {
            return [job.id, { retries: 1, scheduled: now }];
          }
          if (row.retries >= job.maxRetries) {
            logger.error(`Hit max retries; scheduling again ;-; ${job.toString()}`);
            return [job.id, { retries: 0, scheduled: nextSchedule(job.schedule, now) }] as const;
          }
          return [
            job.id,
            {
              retries: row.retries + 1,
              scheduled: now
            }
          ] as const;
        })
      );
      const results = pipe(
        right,
        RA.map(([job]): readonly [string, JobsTableRow] => [
          job.id,
          {
            retries: 0,
            scheduled: nextSchedule(job.schedule, now)
          }
        ]),
        RA.concat(leftResults)
      );

      const newJobsTable = pipe(
        jobsTable,
        RM.toReadonlyArray(S.Ord),
        RA.concat(results),
        RM.fromFoldable(S.Eq, jobsTableRowMagma, RA.Foldable)
      );
      return { lastPingAck: ack, jobsTable: newJobsTable };
    })
  );
